from unittest import expectedFailure


def ExpectedFailure(method):
    """
    Decorator that marks a test method as being expected to fail.

    :param method:  The test method to mark.
    """
    return expectedFailure(method)



import functools

from .. import AbstractTest
from .._constants import IS_TEST_ATTRIBUTE
from .._functions import is_test


def Test(method):
    """
    Decorator which marks a class method as a unit test.

    :param method:  The class method to mark.
    """
    # Abort if the method is already a test
    if is_test(method):
        return method

    # Label the method as a test (will be inherited by wrapper)
    setattr(method, IS_TEST_ATTRIBUTE, True)

    # Wrap the method in the testing infrastructure
    @functools.wraps(method)
    def when_called(test: AbstractTest):
        subject = test.subject()
        resources = test.common_resources()

        if resources is not None:
            return method(test, subject, *resources)
        else:
            return method(test, subject)

    return when_called



from typing import IO

from ._RegressionSerialiser import RegressionSerialiser


class BytesSerialiser(RegressionSerialiser[bytes]):
    """
    Basic serialiser which saves a bytes result in a .bin file.
    """
    @classmethod
    def binary(cls) -> bool:
        return True

    @classmethod
    def extension(cls) -> str:
        return "bin"

    @classmethod
    def serialise(cls, result: bytes, file: IO[bytes]):
        file.write(result)

    @classmethod
    def deserialise(cls, file: IO[bytes]) -> bytes:
        return file.read()



from typing import IO, Optional

from ._RegressionSerialiser import RegressionSerialiser


class StringSerialiser(RegressionSerialiser[str]):
    """
    Basic serialiser which saves a string result in a .txt file.
    """
    @classmethod
    def binary(cls) -> bool:
        return False

    @classmethod
    def extension(cls) -> str:
        return "txt"

    @classmethod
    def serialise(cls, result: str, file: IO[str]):
        file.write(result)

    @classmethod
    def deserialise(cls, file: IO[str]) -> str:
        return file.read()

    @classmethod
    def compare(cls, result: str, reference: str) -> Optional[str]:
        if super().compare(result, reference) is not None:
            return "Result:\n" + \
                result + "\n" + \
                "Reference:\n" + \
                reference



"""
Module for helper functions.
"""
from typing import Any, Dict, Tuple, Optional, Type

from .serialisation import RegressionSerialiser
from . import _constants


def is_test(method) -> bool:
    """
    Checks if the given method is a test method.

    :param method:  The method to check.
    :return:        True if the method is a test method,
                    False if not.
    """
    return getattr(method, _constants.IS_TEST_ATTRIBUTE, False)


def get_skip_reason(method) -> Optional[str]:
    """
    Checks if the given test method should be skipped.

    :param method:  The test method to check.
    :return:        The reason the test was skipped if it was,
                    or None if it wasn't.
    """
    return getattr(method, _constants.SKIP_ATTRIBUTE, None)


def get_subject_args(method) -> Optional[Tuple[Tuple[Any, ...], Dict[str, Any]]]:
    """
    Gets the arguments for the subject initialiser from the given
    method, if any.

    :param method:  The test method to get the arguments from.
    :return:        The arguments, or None if none specified.
    """
    return getattr(method, _constants.SUBJECT_ARGS_ATTRIBUTE, None)


def get_serialisers(method) -> Dict[Type, Type[RegressionSerialiser]]:
    """
    Gets the serialisers for the given test method.

    :param method:  The test method.
    :return:        The mapping from result type to serialiser.
    """
    return getattr(method, _constants.SERIALISERS_ATTRIBUTE, {})



import os
from abc import abstractmethod
from typing import IO, Generic, TypeVar, AnyStr, Optional

# The types of the regression result and the disk type
ResultType = TypeVar("ResultType")


class RegressionSerialiser(Generic[ResultType]):
    """
    Base class for serialisers of regression results to/from disk. Purely
    class-based, should not need to be instantiated.
    """
    @classmethod
    @abstractmethod
    def binary(cls) -> bool:
        """
        Whether this serialiser serialises to binary or text.

        :return:    True if this serialiser class serialises to binary,
                    False if it serialises to text.
        """
        pass

    @classmethod
    @abstractmethod
    def extension(cls) -> str:
        """
        Gets the extension to use for regression files.
        """
        pass

    @classmethod
    def extend(cls, filename: str) -> str:
        """
        Adds the extension to the end of the given filename, if it's not
        already present.

        :param filename:    The filename to extend.
        :return:            The filename, with the extension.
        """
        # Get the extension with a leading dot
        dotted_extension: str = "." + cls.extension()

        # Add the extension if it's not already present
        if not filename.endswith(dotted_extension):
            filename += dotted_extension

        return filename

    @classmethod
    def exists(cls, filename: str) -> bool:
        """
        Whether a regression file exists under the given filename.

        :param filename:    The filename to look for.
        :return:            True if the given regression file exists,
                            False if not.
        """
        return os.path.exists(cls.extend(filename))

    @classmethod
    def save(cls, result: ResultType, filename: str):
        """
        Saves the given result to the given file.

        :param result:      The result to save.
        :param filename:    The name of the file to save to.
        """
        # Add the extension to the filename
        filename = cls.extend(filename)

        # If the path to the file doesn't exist, create it
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        # Determine if we need to open in binary or text mode
        mode = "w" if not cls.binary() else "wb"

        # Serialise the result to the file
        with open(filename, mode) as file:
            cls.serialise(result, file)

    @classmethod
    @abstractmethod
    def serialise(cls, result: ResultType, file: IO[AnyStr]):
        """
        Serialises the given result into the given file.

        :param result:  The result to serialise.
        :param file:    The handle to a file to write to.
        """
        pass

    @classmethod
    def load(cls, filename: str) -> ResultType:
        """
        Loads the regression result from the given file.

        :param filename:    The file to load from.
        :return:            The result stored in the file.
        """
        # Add the extension to the filename
        filename = cls.extend(filename)

        # Determine if we need to open in binary or text mode
        mode = "r" if not cls.binary() else "rb"

        # Deserialise and return the result from the file
        with open(filename, mode) as file:
            return cls.deserialise(file)

    @classmethod
    @abstractmethod
    def deserialise(cls, file: IO[AnyStr]) -> ResultType:
        """
        Deserialises a regression result from the given file.

        :param file:    The file to read from.
        :return:        The regression result stored in the file.
        """
        pass

    @classmethod
    def compare(cls, result: ResultType, reference: ResultType) -> Optional[str]:
        """
        Compares the result to the reference value for equality. Can be
        overridden to define a different sense of equality other than the
        standard == result.

        :param result:      The result of this regression test.
        :param reference:   The reference result from disk.
        :return:            A message describing why the comparison failed,
                            or None if it passed.
        """
        return "result does not equal reference" if result != reference else None



import functools
from typing import Type

from ._Test import Test
from .. import AbstractTest


def ExceptionTest(*exceptions: Type[Exception]):
    """
    Decorator that specifies a test that should raise one
    of the given exception classes.

    :param exceptions:  The exceptions, one of which should be raised.
    """
    def applicator(method):
        # Make the method a test
        method = Test(method)

        # Wrap the method with the exception infrastructure
        @functools.wraps(method)
        def when_called(test: AbstractTest):
            with test.assertRaises(exceptions):
                return method(test)

        return when_called

    return applicator



from typing import Type

from ..serialisation import RegressionSerialiser
from .._constants import SERIALISERS_ATTRIBUTE
from .._functions import get_serialisers


def WithSerialiser(result_type: Type, serialiser: Type[RegressionSerialiser]):
    """
    Decorator for regression tests which tells it to use the given
    serialiser for results of the given type.

    :param result_type:     The type of result the serialiser should be used for.
    :param serialiser:      The serialiser.
    """
    def applicator(method):
        serialisers = get_serialisers(method)

        serialisers.update({
            result_type: serialiser
        })

        setattr(method, SERIALISERS_ATTRIBUTE, serialisers)

        return method

    return applicator



from .._constants import SKIP_ATTRIBUTE


def Skip(reason: str, condition: bool = True):
    """
    Decorator which marks a test as skipped.

    :param reason:      The reason the test was skipped.
    :param condition:   The condition under which to skip the test.
    """
    def applicator(method):
        if condition:
            setattr(method, SKIP_ATTRIBUTE, reason)

        return method

    return applicator



from ._ExceptionTest import ExceptionTest
from ._ExpectedFailure import ExpectedFailure
from ._RegressionTest import RegressionTest
from ._Skip import Skip
from ._SubjectArgs import SubjectArgs
from ._Test import Test
from ._WithSerialiser import WithSerialiser



from abc import ABCMeta
from typing import Dict, Any, Tuple
from unittest import defaultTestLoader, TestCase

from ._functions import is_test

# The method name prefix that unittest looks for when discovering tests
PREFIX: str = defaultTestLoader.testMethodPrefix


class AbstractTestMeta(ABCMeta):
    """
    Meta-class for test classes which allows unittest to discover
    methods marked with one of the test decorators.
    """
    def __new__(mcs, name, bases, namespace, **kwargs):
        # Make sure the test methods get picked up by unittest
        mcs.ensure_tests(namespace)

        # Make sure abstract test classes aren't instantiated by unittest
        bases = mcs.augment_bases(bases, namespace)

        return super().__new__(mcs, name, bases, namespace, **kwargs)

    @staticmethod
    def augment_bases(bases: Tuple, namespace: Dict) -> Tuple:
        """
        Augments the base classes for the new class with the TestCase class.

        :param bases:       The proposed bases for the class.
        :param namespace:   The namespace of the class.
        :return:            The base classes augmented with the TestCase class.
        """
        if not will_be_abstract(bases, namespace):
            if TestCase not in bases:
                return (*bases, TestCase)
            else:
                return bases
        else:
            if TestCase in bases:
                return tuple(base for base in bases if base is not TestCase)
            else:
                return bases

    @staticmethod
    def ensure_tests(namespace: Dict[str, Any]):
        """
        Makes sure that all methods marked with the test tag will be picked up by unittest.

        :param namespace:   The namespace of the new class.
        :return:            The namespace augmented with the unit-test methods.
        """
        # Required so namespace doesn't change during iteration
        new_attributes = {}

        # Check all attributes in the namespace
        for name, attribute in namespace.items():
            if is_test(attribute) and not name.startswith(PREFIX):
                # Create the prefixed name for the test
                test_name = PREFIX + name

                # If there is already an attribute with that name, it's fatal
                if test_name in namespace:
                    raise NameError("Test method '" +
                                    name +
                                    "' clashes with a pre-existing method '" +
                                    test_name +
                                    "'")

                # Add the prefixed name to the namespace
                new_attributes[test_name] = attribute

        # Add the prefixed names to the class attribute namespace
        namespace.update(new_attributes)


# The following code was copied here from wai.common, as we can't
# create a dependency on wai.common without creating a dependency
# cycle (wai.common depends on wai.test for testing).


def get_abstract_methods(bases: Tuple, namespace: Dict[str, Any]):
    """
    Gets the names of the abstract methods that will result from
    a class created with the given base-class set and namespace.

    :param bases:       The base-class set.
    :param namespace:   The namespace.
    :return:            The set of abstract method names.
    """
    abstract_methods = set()

    for base in bases:
        abstract_methods.update(getattr(base, '__abstractmethods__', set()))

    for name, value in namespace.items():
        if is_abstract_function(value):
            abstract_methods.add(name)
        else:
            if name in abstract_methods:
                abstract_methods.remove(name)

    return abstract_methods


def will_be_abstract(bases: Tuple, namespace: Dict[str, Any]):
    """
    Determines if a class made with the given set of base-classes
    and namespace will be abstract or concrete.

    :param bases:       The set of base classes.
    :param namespace:   The namespace of the new class.
    :return:            True if the newly-created class will be abstract,
                        False if it will be concrete.
    """
    return len(get_abstract_methods(bases, namespace)) > 0


def is_abstract_function(func):
    """
    Whether the given function is abstract in its class.

    :param func:    The function to check.
    :return:        True if the function is abstract,
                    False if not.
    """
    return getattr(func, '__isabstractmethod__', False)



from .._constants import SUBJECT_ARGS_ATTRIBUTE


def SubjectArgs(*args, **kwargs):
    """
    Decorator that overrides the default arguments to the subject
    initialiser.

    :param args:        The positional arguments to the subject initialiser.
    :param kwargs:      The keyword arguments to the subject initialiser.
    """
    def applicator(method):
        setattr(method, SUBJECT_ARGS_ATTRIBUTE, (args, kwargs))

        return method

    return applicator



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# setup.py
# Copyright (C) 2018 Fracpete (fracpete at waikato dot ac dot nz)

from setuptools import setup, find_namespace_packages


def _read(f) -> bytes:
    """
    Reads in the content of the file.
    :param f: the file to read
    :type f: str
    :return: the content
    :rtype: str
    """
    return open(f, 'rb').read()


setup(
    name="wai.test",
    description="Python library for unit tests.",
    long_description=(
        _read('DESCRIPTION.rst') + b'\n' +
        _read('CHANGES.rst')).decode('utf-8'),
    url="https://github.com/waikato-datamining/wai-test",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Programming Language :: Python :: 3',
    ],
    license='MIT',
    package_dir={
        '': 'src'
    },
    packages=find_namespace_packages(where="src"),
    namespace_packages=[
        "wai"
    ],
    version="0.0.3",
    author='Corey Sterling',
    author_email='coreytsterling@gmail.com',
    install_requires=[],
    include_package_data=True
)



"""
Module for constants used by the wai.test library.
"""
# Attribute of methods that are to be run as tests
IS_TEST_ATTRIBUTE: str = "__test"

# Attribute of test methods which override the default subject initialiser
SUBJECT_ARGS_ATTRIBUTE: str = "__subject_args"

# Attribute of test methods indicating they should be skipped
SKIP_ATTRIBUTE: str = "__skip"

# Attribute of serialisers that the test method will use
SERIALISERS_ATTRIBUTE: str = "__serialisers"



import inspect
import os
from abc import abstractmethod
from typing import Any, Dict, Tuple, Optional, Type
from unittest import TestCase

from ._AbstractTestMeta import AbstractTestMeta
from .serialisation import BytesSerialiser, StringSerialiser, RegressionSerialiser
from ._functions import get_subject_args, get_skip_reason, get_serialisers, is_test

# The default location to store regression test results
DEFAULT_REGRESSION_ROOT = os.path.join(".", "resources", "regression")


class AbstractTest(TestCase, metaclass=AbstractTestMeta):
    def __init__(self, methodName='runTest'):
        # Get the method this instance is testing
        method = getattr(self, methodName)

        # If it's a marked test, it may have been given a
        # prefix, so substitute the base method's name
        if is_test(method):
            methodName = method.__name__

        super().__init__(methodName)

    @classmethod
    @abstractmethod
    def subject_type(cls):
        """
        Defines the subject callable for this test case.
        """
        pass

    @classmethod
    def common_resources(cls) -> Optional[Tuple[Any, ...]]:
        """
        Loads the common resources for all tests in this class.
        By default there are no common resources.
        """
        return None

    @classmethod
    def common_serialisers(cls) -> Optional[Dict[Type, Type[RegressionSerialiser]]]:
        """
        Defines additional common serialisers to use for regression
        tests in this class (in addition to the basic string and bytes
        serialisers, although these can be overridden).
        """
        return None

    @classmethod
    def common_arguments(cls) -> Optional[Tuple[Tuple[Any, ...], Dict[str, Any]]]:
        """
        Defines the default arguments to pass to the subject on instantiation.
        By default, no arguments are passed.
        """
        return None

    def setUp(self) -> None:
        # Get the method being tested
        test_method = self.get_test_method()

        # If the method is marked with skip, skip it
        skip_reason = get_skip_reason(test_method)
        if skip_reason is not None:
            self.skipTest(skip_reason)

    @classmethod
    def instantiate_subject(cls, *args, **kwargs) -> Any:
        """
        Instantiates a test subject with the given arguments.
        """
        return cls.subject_type()(*args, **kwargs)

    def subject(self) -> Any:
        # Get the arguments to use to create the test subject from the test method
        subject_args = get_subject_args(self.get_test_method())

        # If there are arguments, use them
        if subject_args is not None:
            return self.instantiate_subject(*subject_args[0], **subject_args[1])

        # If not, try using the default arguments
        subject_args = self.common_arguments()

        # If there are default arguments, use them
        if subject_args is not None:
            return self.instantiate_subject(*subject_args[0], **subject_args[1])

        # Otherwise use the default constructor
        return self.instantiate_subject()

    def handle_regression_results(self, results):
        """
        Handles comparing the results of a regression test to
        the results stored in the regression files.

        :param results:     The results of the regression test.
        """
        # Make sure the regression results are a named map
        if not isinstance(results, dict) or not all(isinstance(key, str) for key in results):
            self.fail("Regression test '" +
                      self.get_test_method_name() +
                      "' didn't return a named map of regression results")

        # Handle each result individually
        for name, result in results.items():
            self.handle_regression_result(name, result)

    def handle_regression_result(self, name: str, result: Any):
        """
        Handles the comparison of an individual regression result
        to the stored reference.

        :param name:    The name of the regression to use.
        :param result:  The result of the regression test.
        """
        # Treat each individual regression result as a sub-test
        with self.subTest(regression=name):
            # Get the serialisers for this test method
            serialisers = self.get_serialisers()

            # Get the preference order of serialiser types
            mro = inspect.getmro(type(result))

            # Get a serialiser for this result type
            serialiser = None
            for result_type in mro:
                if result_type in serialisers:
                    serialiser = serialisers[result_type]
                    break

            # Make sure we have a serialiser
            if serialiser is None:
                self.fail("No regression serialiser found for result of type: " + type(result).__name__)

            # Complete the file path for this result
            filename: str = os.path.join(self.get_regression_path(), self.get_test_method_name(), name)

            # If the regression file doesn't exist yet, create it
            if not serialiser.exists(filename):
                serialiser.save(result, filename)

            # Otherwise load and check the saved result
            else:
                reference = serialiser.load(filename)

                # Use the serialiser's notion of equality
                failure_message: str = serialiser.compare(result, reference)
                if failure_message is not None:
                    self.fail(serialiser.__name__ + ": " + failure_message)

    @classmethod
    def get_regression_root_path(cls) -> str:
        """
        Gets the path to the root directory to store regression results in.
        Can be overridden for a custom directory per test-class.

        :return:    The path to the root directory.
        """
        return DEFAULT_REGRESSION_ROOT

    @classmethod
    def get_regression_path(cls) -> str:
        """
        Gets the path to the regression results for this test-class.

        :return:    The path.
        """
        return os.path.join(cls.get_regression_root_path(), cls.get_relative_regression_path())

    @classmethod
    def get_relative_regression_path(cls) -> str:
        """
        Gets the path to the regression results for this test-class,
        relative to the root path.

        :return:    The path.
        """
        # Get the fully-qualified name of the subject (in dotted form)
        fully_qualified_name: str = cls.subject_type().__module__ + '.' + cls.subject_type().__qualname__

        # Replace the dots with platform-dependent slashes
        return fully_qualified_name.replace(".", os.sep)

    def get_serialisers(self) -> Dict[Type, Type[RegressionSerialiser]]:
        """
        Gets the serialisers to use for this test method.

        :return:    The map from type to serialiser.
        """
        # Define the basic serialisers
        serialisers = {
            str: StringSerialiser,
            bytes: BytesSerialiser
        }

        # Add the common serialisers for this class
        common_serialisers = self.common_serialisers()
        if common_serialisers is not None:
            serialisers.update(common_serialisers)

        # Override with any test-specific serialisers
        serialisers.update(get_serialisers(self.get_test_method()))

        return serialisers

    def get_test_method(self):
        """
        Gets the test method being tested by this test instance.

        :return:    The test method.
        """
        return getattr(self, self.get_test_method_name())

    def get_test_method_name(self) -> str:
        """
        Gets the name of the test method being tested by this test instance.

        :return:    The name of the test method.
        """
        return self._testMethodName



import functools

from ._Test import Test
from .. import AbstractTest


def RegressionTest(method):
    """
    Decorator which marks a class method as a regression test.
    """
    # Make the method a test
    method = Test(method)

    # Wrap the method with the regression infrastructure
    @functools.wraps(method)
    def when_called(test: AbstractTest):
        results = method(test)

        test.handle_regression_results(results)

    return when_called
